/***************************************************************************
                          kstars.cpp  -  K Desktop Planetarium
                             -------------------
    begin                : Mon Feb  5 01:11:45 PST 2001
    copyright            : (C) 2001 by Jason Harris
    email                : jharris@30doradus.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qfont.h>
#include <qpopupmenu.h>
#include <qtextstream.h>
#include <qlineedit.h>
#include <stdio.h>
#include <stream.h>
#include <kapp.h>
#include <kconfig.h>
#include <kaccel.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <qlayout.h>

#include "kstars.h"

KStars::KStars() : KMainWindow( NULL, NULL )
{
	resize( 640, 600 );
	initMenuBar();
	initStatusBar();

	KDirs = new KStandardDirs();

	geoList.setAutoDelete( TRUE );
	starList.setAutoDelete( TRUE );
	messList.setAutoDelete( TRUE );
	ngcList.setAutoDelete( TRUE );
	icList.setAutoDelete( TRUE );
	clineList.setAutoDelete( TRUE );
	clineModeList.setAutoDelete( TRUE );
	cnameList.setAutoDelete( TRUE );

	// Get initial Location from config()
	kapp->config()->setGroup( "Location" );
	QString city = kapp->config()->readEntry( "City", "Baltimore" );
	QString state = kapp->config()->readEntry( "State", "Maryland" );
	kapp->config()->setGroup( "View" );
	SkyColor = kapp->config()->readEntry( "SkyColor", "#002" );
	StarColor = kapp->config()->readEntry( "StarColor", "#FFF" );
	EqColor = kapp->config()->readEntry( "EqColor", "#FFF" );
	HorzColor = kapp->config()->readEntry( "HorzColor", "#5A3" );
	MessColor = kapp->config()->readEntry( "MessColor", "#0F0" );
	NGCColor = kapp->config()->readEntry( "NGCColor", "#066" );
	ICColor = kapp->config()->readEntry( "ICColor", "#439" );
	CLineColor = kapp->config()->readEntry( "CLineColor", "#555" );
	CNameColor = kapp->config()->readEntry( "CNameColor", "#AA7" );
	HSTColor = kapp->config()->readEntry( "HSTColor", "#A00" );
	drawBSC = kapp->config()->readBoolEntry( "ShowBSC", true );
	drawMessier = kapp->config()->readBoolEntry( "ShowMess", true );
	drawNGC = kapp->config()->readBoolEntry( "ShowNGC", false );
	drawIC = kapp->config()->readBoolEntry( "ShowIC", false );
	drawConstellLines = kapp->config()->readBoolEntry( "ShowCLines", true );
	drawConstellNames = kapp->config()->readBoolEntry( "ShowCNames", false );
	drawEquator = kapp->config()->readBoolEntry( "ShowEquator", true );
	drawEcliptic = kapp->config()->readBoolEntry( "ShowEcliptic", true );
	drawHorizon = kapp->config()->readBoolEntry( "ShowHorizon", true );
	drawGround = kapp->config()->readBoolEntry( "ShowGround", true );
	drawSun = kapp->config()->readBoolEntry( "ShowSun", true );
	drawMoon = kapp->config()->readBoolEntry( "ShowMoon", true );
  useAltAz = kapp->config()->readBoolEntry( "UseAltAz", false );

  //Initialize object type strings
	TypeName[0] = "star";
	TypeName[1] = "double star";
	TypeName[2] = "planet";
	TypeName[3] = "open cluster";
	TypeName[4] = "globular cluster";
	TypeName[5] = "gaseous nebula";
	TypeName[6] = "planetary nebula";
	TypeName[7] = "supernova remnant";
	TypeName[8] = "galaxy";
	TypeName[9] = "UNUSED_TYPE";
	
	//Initialize location
	readCityData();
	
	unsigned int i;
	for ( i=0; i<geoList.count(); ++i ) {
		if ( QString( geoList.at(i)->name() ) == city && QString( geoList.at(i)->state() ) == state ) break;
	}
	geo = new GeoLocation ( *geoList.at(i) );

	//start the clock
	lastSecond = QTime::currentTime().second();
	last_msec = QTime::currentTime().msec();

	LTime = new QDateTime( QDate::currentDate(), QTime::currentTime() );
	UTime = new QDateTime( *LTime );
	*UTime = UTime->addSecs( geo->TZ()*3600 );

  LST = new QTime();
  *LST = UTtoLST( *UTime, geo->lng() );

  tmr = new QTimer( this );
	tmr->start( TIMER_INTERVAL, FALSE );
	QObject::connect( tmr, SIGNAL( timeout() ), this, SLOT( updateTime() ) );

	skymap = new SkyMap( this );
//	skymap->setGeometry( 0, 90, 640, 500 );
	skymap->setMinimumSize( 640, 500 );
	skymap->setMaximumSize( 640, 500 );
	skymap->LSTh.setD( LST->hour(), LST->minute(), LST->second() );
	skymap->HourAngle = skymap->LSTh.getD() - skymap->focus.getRA();
	skymap->focus.RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
  skymap->oldfocus.setAz( skymap->focus.getAz() );
  skymap->oldfocus.setAlt( skymap->focus.getAlt() );

	//Read in sky object data
	readStarData();
	readNGCData();
	readURLData();
	readCLineData();
	readCNameData();

	//create dialogs
	timedialog = new TimeDialog( this );
	locationdialog = new LocationDialog( this );
	finddialog = new FindDialog( this );
	viewopsdialog = new ViewOpsDialog( this );

	//Define the Celestial Equator
	for ( unsigned int i=0; i<720; ++i ) {
		SkyPoint *o = new SkyPoint( i/30., 0.0 );
		o->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
		Equator.append( o );
	}

  // Define the horizon.
  // Use the celestial Equator as a convenient starting point, but instead of RA and Dec,
  // interpret the coordinates as azimuth and altitude, and then convert to RA, dec.
  // The horizon will be redefined whenever the positions of sky objects are updated.
	for ( unsigned int i=0; i<Equator.count(); ++i ) {
		double dec, sinlat, coslat, sindec, cosdec;
		double HA, RA, Az, sinAz, cosAz;
		Az = HourToDeg( Equator.at(i)->getRA() );
		SinCos( Az, sinAz, cosAz );
		SinCos( geo->lat(), sinlat, coslat );

		dec = RadToDeg( asin( coslat*cosAz ) );
		SinCos( dec, sindec, cosdec );
   	HA = RadToHour( acos( -1.0*(sinlat*sindec)/(coslat*cosdec) ) );
		if ( sinAz > 0.0 ) { HA = 24.0 - HA; }
		RA = skymap->LSTh.getD() - HA;

		SkyPoint *o = new SkyPoint( RA, dec );
		o->setAlt( 0.0 );
    o->setAz( Az );

		Horizon.append( o );
	}

	//Create main vertical layout manager
	vlay = new QVBoxLayout( this );
		
	//Create the infoPanel
	infoPanel = new QFrame( this );
	infoPanel->setFrameShape( QFrame::Panel );
	
	//Create main Layout manager for infoPanel
	iplay = new QHBoxLayout( infoPanel, 2, 2, "iplay" );
	
	//Create Widgets to appear in the info panel
  LTLabel = new QLabel( i18n( "LT:" ), infoPanel );
  UTLabel = new QLabel( i18n( "UT:" ), infoPanel );
	STLabel = new QLabel( i18n( "ST:" ), infoPanel );
	JDLabel = new QLabel( i18n( "JD:" ), infoPanel );
  LT = new QLabel( LTime->time().toString(), infoPanel );
  UT = new QLabel( UTime->time().toString(), infoPanel );
  ST = new QLabel( "00:00:00", infoPanel );
	double jdate = getJD( UTime );
  JD = new QLabel( QString("%1").arg(jdate, 10, 'f', 2), infoPanel );
	LTDate = new QLabel( getDateString(LTime->date() ), infoPanel );
	UTDate = new QLabel( getDateString(UTime->date() ), infoPanel );
	CityName = new QLabel( QString( geo->name() ), infoPanel );
	StateName = new QLabel( QString( geo->state() ), infoPanel );
	LongLabel = new QLabel( i18n( "Long: " ), infoPanel );
	LatLabel = new QLabel( i18n( "Lat:  " ), infoPanel );
	Long = new QLabel( QString::number( geo->lng(), 'f', 3 ), infoPanel );
	Long->setAlignment( AlignRight );
	Lat = new QLabel( QString::number( geo->lat(), 'f', 3 ), infoPanel );
	Lat->setAlignment( AlignRight );

	//create layouts to be added to iP_hlay
	tlablay = new QVBoxLayout( 2, "tlablay" );
	timelay = new QVBoxLayout( 2, "timelay" );
	datelay = new QVBoxLayout( 2, "datelay" );
	geolay = new QVBoxLayout( 2, "geolay" );
	coolay = new QGridLayout( 2, 2, 2, "coolay" );
	
	QSpacerItem *tlabspacer = new QSpacerItem( 10, 10, QSizePolicy::Minimum, QSizePolicy::Expanding );
	QSpacerItem *timespacer = new QSpacerItem( 10, 10, QSizePolicy::Minimum, QSizePolicy::Expanding );
	QSpacerItem *datespacer = new QSpacerItem( 10, 10, QSizePolicy::Minimum, QSizePolicy::Expanding );
	QSpacerItem *geospacer = new QSpacerItem( 10, 10, QSizePolicy::Minimum, QSizePolicy::Expanding );
	
	//Pack widgets into infoPanel
	tlablay->addWidget( LTLabel );
	tlablay->addWidget( UTLabel );
	tlablay->addWidget( STLabel );
	tlablay->addWidget( JDLabel );
	tlablay->addItem( tlabspacer );
	
	timelay->addWidget( LT );
	timelay->addWidget( UT );
	timelay->addWidget( ST );
	timelay->addWidget( JD );
	timelay->addItem( timespacer );
	
	datelay->addWidget( LTDate );
	datelay->addWidget( UTDate );
	datelay->addItem( datespacer );

	geolay->addWidget( CityName );
	geolay->addWidget( StateName );
	
	coolay->addWidget( LongLabel, 0, 0 );
	coolay->addWidget( LatLabel, 1, 0 );
	coolay->addWidget( Long, 0, 1 );
	coolay->addWidget( Lat, 1, 1 );
	
	geolay->addLayout( coolay );
	geolay->addItem( geospacer );
		
	iplay->addLayout( tlablay );
	iplay->addSpacing( 6 );
	iplay->addLayout( timelay );
	iplay->addSpacing( 12 );
	iplay->addLayout( datelay );
	iplay->addSpacing( 12 );
	iplay->addLayout( geolay );
	
//	QSpacerItem *ipspacer = new QSpacerItem( 10, 10, QSizePolicy::Minimum, QSizePolicy::Expanding );
//	iplay->addItem( ipspacer );
	iplay->addStretch();
	iplay->activate();
	
	vlay->addSpacing( menuBar()->height() );
	vlay->addWidget( infoPanel );
	vlay->addWidget( skymap );
	vlay->addWidget( statusBar() );
	vlay->activate();
}

KStars::~KStars()
{
	//Sync the config file
	kapp->config()->setGroup( "Location" );
	kapp->config()->writeEntry( "City", CityName->text() );
	kapp->config()->writeEntry( "State", StateName->text() );
	kapp->config()->setGroup( "View" );
	kapp->config()->writeEntry( "SkyColor", SkyColor );
	kapp->config()->writeEntry( "StarColor", StarColor );
	kapp->config()->writeEntry( "EqColor", EqColor );
	kapp->config()->writeEntry( "HorzColor", HorzColor );
	kapp->config()->writeEntry( "MessColor", MessColor );
	kapp->config()->writeEntry( "NGCColor", NGCColor );
	kapp->config()->writeEntry( "ICColor", ICColor );
	kapp->config()->writeEntry( "CLineColor", CLineColor );
	kapp->config()->writeEntry( "CNameColor", CNameColor );
	kapp->config()->writeEntry( "HSTColor", HSTColor );
	kapp->config()->writeEntry( "ShowBSC", drawBSC );
	kapp->config()->writeEntry( "ShowMess", drawMessier );
	kapp->config()->writeEntry( "ShowNGC", drawNGC );
	kapp->config()->writeEntry( "ShowIC", drawIC );
	kapp->config()->writeEntry( "ShowCLines", drawConstellLines );
	kapp->config()->writeEntry( "ShowCNames", drawConstellNames );
	kapp->config()->writeEntry( "ShowEquator", drawEquator );
	kapp->config()->writeEntry( "ShowEcliptic", drawEcliptic );
	kapp->config()->writeEntry( "ShowHorizon", drawHorizon );
	kapp->config()->writeEntry( "ShowGround", drawGround );
	kapp->config()->writeEntry( "ShowSun", drawSun );
	kapp->config()->writeEntry( "ShowMoon", drawMoon );
	kapp->config()->writeEntry( "UseAltAz", useAltAz );
	
	kapp->config()->writeEntry( "PixelScale", skymap->pixelScale );
	kapp->config()->writeEntry( "FocusRA", skymap->focus.getRA() );
	kapp->config()->writeEntry( "FocusDec", skymap->focus.getDec() );
	kapp->config()->sync();
	
	// remove entries in all lists:
  // geoList, starList, messList, ngcList, icList, Horizon, Equator,
  // clineList, cnameList, clineModeList
	for ( unsigned int i=0; i<geoList.count(); ++i ) {
		geoList.remove( i );
	}

	for ( unsigned int i=0; i<starList.count(); ++i ) {
		starList.remove( i );
	}
	
	for ( unsigned int i=0; i<messList.count(); ++i ) {
		messList.remove( i );
	}

	for ( unsigned int i=0; i<ngcList.count(); ++i ) {
		ngcList.remove( i );
	}

	for ( unsigned int i=0; i<icList.count(); ++i ) {
		icList.remove( i );
	}

	for ( unsigned int i=0; i<Horizon.count(); ++i ) {
		Horizon.remove( i );
	}

	for ( unsigned int i=0; i<Equator.count(); ++i ) {
		Equator.remove( i );
	}

	for ( unsigned int i=0; i<clineList.count(); ++i ) {
		clineList.remove( i );
	}

	for ( unsigned int i=0; i<cnameList.count(); ++i ) {
		cnameList.remove( i );
	}

	for ( unsigned int i=0; i<clineModeList.count(); ++i ) {
		clineModeList.remove( i );
	}

	// delete dialog window objects
	delete timedialog;
	delete locationdialog;
	delete viewopsdialog;
	
	delete geo;
	delete LT;
	delete UT;
	delete ST;
	delete JD;
	delete LTDate;
	delete UTDate;
	delete LTLabel;
	delete UTLabel;
	delete STLabel;
	delete JDLabel;
	delete LTime;
	delete UTime;
	delete LST;
	delete CityName;
	delete StateName;
	delete LongLabel;
	delete LatLabel;
	
	delete skymap;
}

void KStars::initMenuBar() {
	QPopupMenu *p;
	int id;

	KAccel *a = new KAccel( this );
	a->connectItem( KStdAccel::Quit, this, SLOT( close() ) );
	
	p = new QPopupMenu;
	id = p->insertItem( i18n( "&Quit" ), this, SLOT( close() ));
	a->changeMenuAccel( p, id, KStdAccel::Quit );
	menuBar()->insertItem( i18n( "&File" ), p );

	p = new QPopupMenu;
	p->insertItem( i18n( "&Now" ), this, SLOT( mSetTimeToNow() ));
	p->insertItem( i18n( "&Set..." ), this, SLOT( mSetTime() ));
	p->insertItem( i18n( "&Run/Stop" ), this, SLOT( mToggleTimer() ));
	menuBar()->insertItem( i18n( "&Time" ), p );

	p = new QPopupMenu;
	p->insertItem( i18n( "&Zenith" ), this, SLOT( mZenith() ));
	int findItemID = p->insertItem( i18n( "&Find" ), this, SLOT( mFind() ));
//	p->setItemEnabled( findItemID, false );
	int trackItemID = p->insertItem( i18n( "&Track" ), this, SLOT( mTrack() ));
	p->setItemEnabled( trackItemID, false );
	p->insertSeparator();
	p->insertItem( i18n( "&Geographic..." ), this, SLOT( mGeoLocator() ));
	menuBar()->insertItem( i18n( "&Location" ), p );

	p = new QPopupMenu;
	p->insertItem( i18n( "&View Options..." ), this, SLOT( mViewOps() ));	
	p->insertItem( i18n( "&Reverse Video" ), this, SLOT( mReverseVideo() ));
	p->insertItem( i18n( "Zoom &In" ), this, SLOT( mZoomIn() ));
	p->insertItem( i18n( "Zoom &Out" ), this, SLOT( mZoomOut() ));
	menuBar()->insertItem( i18n( "&View" ), p );

	menuBar()->insertItem( i18n( "&Help" ), helpMenu() );

}

void KStars::initStatusBar() {
	statusBar()->insertItem( i18n( " Welcome to KStars " ), 0, 1, true );
	statusBar()->setItemAlignment( 0, AlignLeft | AlignVCenter );
	CurrentPosition = "     00:00:00, +00:00:00 ";       	

	statusBar()->insertItem( CurrentPosition, 1, 1, true );
	statusBar()->setItemAlignment( 1, AlignRight | AlignVCenter );
	statusBar()->setItemFixed( 1, -1 );
}

/** Menu Slot Functions **/
void KStars::mSetTimeToNow() {
	LTime->setTime( QTime::currentTime() );
	LTime->setDate( QDate::currentDate() );
	UTime->setTime( LTime->time() );
	UTime->setDate( LTime->date() );
	*UTime = UTime->addSecs( geo->TZ()*3600 );

	lastSecond = QTime::currentTime().second();
	updateTime();
}

void KStars::mSetTime() {
	if (tmr->isActive() ) { tmr->stop(); }

	if ( timedialog->exec() == QDialog::Accepted ) {
		QTime newTime( timedialog->HourBox->value(),
					   timedialog->MinuteBox->value(),
					   timedialog->SecondBox->value() );

		QDate newDate( timedialog->dPicker->getDate() );

		LTime->setTime( newTime );
		LTime->setDate( newDate );
		UTime->setTime( newTime );
		UTime->setDate( newDate );
		*UTime = UTime->addSecs( geo->TZ()*3600 );

		lastSecond = QTime::currentTime().second();
		updateTime();
	}
}

void KStars::mToggleTimer() {
	if ( tmr->isActive() ) {
		tmr->stop();
	} else {
		tmr->start( TIMER_INTERVAL, FALSE );
	}
}

void KStars::mZenith() {
	skymap->focus.setAlt( 90.0 );
	skymap->focus.AltAzToRADec( skymap->LSTh.getD(), geo->lat() );
	skymap->HourAngle = skymap->LSTh.getD() - skymap->focus.getRA();
	skymap->update();
}

void KStars::mZoomIn() {
	if ( skymap->pixelScale < 16384 ) {
		skymap->pixelScale *= 2;
		//If using Alt-Az, update the Alt-Az coordinates
		if ( useAltAz && skymap->pixelScale > 10000 ) {
			oldfocus.set( oldfocus.getRA(), oldfocus.getDec()+5.0 );
			updateTime();
		}
		skymap->update();
	}
}

void KStars::mZoomOut() {
	if ( skymap->pixelScale > 256 ) {
		skymap->pixelScale /= 2;
		skymap->update();
	}
}

void KStars::mFind() {
	if ( finddialog->exec() == QDialog::Accepted ) {
		SkyObject newObj = (SkyObject)*finddialog->filteredList.at( finddialog->SearchList->currentItem() );
		skymap->focus.set( newObj.getRA(), newObj.getDec() );
		skymap->HourAngle = skymap->LSTh.getD() - skymap->focus.getRA();
		skymap->update();
	}
}

void KStars::mTrack() {
}

void KStars::mViewOps() {
	QString oldStarColor = StarColor;
	QString oldSkyColor = SkyColor;
	QString oldMessColor = MessColor;
	QString oldNGCColor = NGCColor;
	QString oldICColor = ICColor;
	QString oldCNameColor = CNameColor;
	QString oldCLineColor = CLineColor;
	QString oldEqColor = EqColor;
	QString oldHorzColor = HorzColor;
	QString oldHSTColor = HSTColor;
	
	if ( viewopsdialog->exec() != QDialog::Accepted ) {
		//User pressed cancel; revert colors
		StarColor = oldStarColor;
		SkyColor = oldSkyColor;
		MessColor = oldMessColor;
		NGCColor = oldNGCColor;
		ICColor = oldICColor;
		CNameColor = oldCNameColor;
		CLineColor = oldCLineColor;
		EqColor = oldEqColor;
		HorzColor = oldHorzColor;
		HSTColor = oldHSTColor;
	}
}

void KStars::mReverseVideo() {
	QString temp = SkyColor;
	SkyColor = StarColor;
	StarColor = temp;
	skymap->update();
}


void KStars::mGeoLocator() {
	if ( locationdialog->exec() == QDialog::Accepted ) {
 		if ( locationdialog->NewLong->text().isEmpty() ||
	 			locationdialog->NewLat->text().isEmpty() ) {
	 		int ii = locationdialog->getCityIndex();
 			if ( ii >= 0 ) {
 				geo->reset( geoList.at(ii) );
 				CityName->setText( geo->name() );
	 			StateName->setText( geo->state() );
 				Long->setText( QString::number( geo->lng(), 'f', 3 ) );
 				Lat->setText( QString::number( geo->lat(), 'f', 3 ) );

 				// Adjust Local time for new time zone
	 			LTime->setDate( UTime->date() );
 				LTime->setTime( UTime->time() );
 				*LTime = LTime->addSecs( geo->TZ()*-3600 );
	 		}
 		} else {
 			geo->setLong( locationdialog->NewLong->text().toFloat() );
 			geo->setLat( locationdialog->NewLat->text().toFloat() );
 			geo->setName( locationdialog->NewCityName->text().latin1() );
 			geo->setState( locationdialog->NewStateName->text().latin1() );
 			Long->setText( QString::number( geo->lng(), 'f', 3 ) );
 			Lat->setText( QString::number( geo->lat(), 'f', 3 ) );
 			CityName->setText( geo->name() );
	 		StateName->setText( geo->state() );
 		
	 		locationdialog->NewLong->setText( "" );
	 		locationdialog->NewLat->setText( "" );
	 		locationdialog->NewCityName->setText( "" );
	 		locationdialog->NewStateName->setText( "" );
	 		
 		}
 	}
}

void KStars::updateTime( void ) {
  thisSecond = QTime::currentTime().second();
  this_msec = QTime::currentTime().msec();

  if ( thisSecond != lastSecond ) {
	  lastSecond = thisSecond;
	  last_msec = this_msec;

	  *LTime = LTime->addSecs( 1 );
	  *UTime = UTime->addSecs( 1 );
	}

	LT->setText( LTime->time().toString() );
	UT->setText( UTime->time().toString() );
	LTDate->setText( getDateString( LTime->date() ) );
	UTDate->setText( getDateString( UTime->date() ) );

	double jdate = getJD( UTime );
	JD->setText( QString("%1").arg(jdate, 10, 'f', 2) );

  QTime t = UTime->time().addMSecs( this_msec - last_msec );
  QDateTime date( UTime->date(), t );
  *LST = UTtoLST( date, geo->lng() );

	QString dummy;
	QString STString = dummy.sprintf( "%02d:%02d:%02d",
			  LST->hour(), LST->minute(), LST->second() );
	ST->setText( STString );

	skymap->LSTh.setD( LST->hour(), LST->minute(), LST->second() );

	skymap->focus.setRA( skymap->LSTh.getD() - skymap->HourAngle );
	skymap->focus.RADecToAltAz( skymap->LSTh.getD(), geo->lat() );

  // Update positions of objects, if necessary
	if ( fabs( skymap->focus.getRA() - skymap->oldfocus.getRA()) > 10./skymap->pixelScale ) {
		//Recompute Alt, Az coords for all objects
		//Stars
		if (drawBSC) {
			for ( unsigned int i=0; i<starList.count(); ++i ) {
				starList.at(i)->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
			}
		}
		
		//Messier
		if (drawMessier) {
			for ( unsigned int i=0; i<messList.count(); ++i ) {
				messList.at(i)->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
			}
		}
		
		//NGC
		if (drawNGC) {
			for ( unsigned int i=0; i<ngcList.count(); ++i ) {
				ngcList.at(i)->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
			}
		}
		
		//IC
		if (drawIC) {
			for ( unsigned int i=0; i<icList.count(); ++i ) {
				icList.at(i)->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
			}
		}
		
		//CLines
		if (drawConstellLines) {
			for ( unsigned int i=0; i<clineList.count(); ++i) {
				clineList.at(i)->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
			}
		}
	
		//CNames
		if (drawConstellNames) {
			for ( unsigned int i=0; i<cnameList.count(); ++i ) {
				cnameList.at(i)->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
			}
		}
	
		//Celestial Equator
		if (drawEquator) {
			for ( unsigned int i=0; i<Equator.count(); ++i ) {
				Equator.at(i)->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
			}
		}
		
		//Horizon: differnet than the others; Alt & Az remain constant, RA, Dec must keep up
		if (drawHorizon || drawGround) {
			for ( unsigned int i=0; i<Horizon.count(); ++i ) {
				Horizon.at(i)->AltAzToRADec( skymap->LSTh.getD(), geo->lat() );
			}
		}

		skymap->oldfocus.set( skymap->focus.getRA(), skymap->focus.getDec() );
		skymap->oldfocus.RADecToAltAz( skymap->LSTh.getD(), geo->lat() );

		skymap->update();
	}

}

QString KStars::getDateString( QDate date ) {
  QString dummy;
  QString dateString = dummy.sprintf( "%02d / %02d / %04d",
			  date.month(), date.day(), date.year() );
  return dateString;
}

double KStars::getJD( QDateTime *t ) {
  int year = t->date().year();
  int month = t->date().month();
  int day = t->date().day();
  int hour = t->time().hour();
  int minute = t->time().minute();
  int second = t->time().second();
  int m, y, A, B, C, D;

  if (month <= 2) {
	  m = month;
	  y = year;
  } else {
	  y = year - 1;
	  m = month + 12;
  }

 /*  If the date is after 10/15/1582, then take Pope Gregory's modification
	 to the Julian calendar into account */

	 if (( year  >1582 ) ||
		 ( year ==1582 && month  >9 ) ||
		 ( year ==1582 && month ==9 && day >15 ))
	 {
		 A = int(y/100);
		 B = 2 - A + int(A/4);
	 } else {
		 B = 0;
	 }

  if (y < 0) {
	  C = int((365.25*y) - 0.75);
  } else {
	  C = int(365.25*y);
  }

  D = int(30.6001*(m+1));

  float d = float(day) + (float(hour) + (float(minute) + second/60.0)/60.0)/24.0;

  double jd = B + C + D + d + 1720994.5;

  return jd;
}

QTime KStars::UTtoLST( QDateTime UT, float longitude) {
  QTime GST = UTtoGST( UT );
  QTime LST = GSTtoLST( GST, longitude );
  return LST;
}

QTime KStars::UTtoGST( QDateTime UT ) {
  QDateTime t0 = UT;
  t0.setTime( QTime(0,0,0) );
  double jd0 = getJD( &t0 );
  double s = jd0 - 2451545.0;
  double t = s/36525.0;
  double t1 = 6.697374558 + 2400.051336*t + 0.000025862*t*t;

	 while (t1 >= 24.0) {t1 -= 24.0;}
	 while (t1 <   0.0) {t1 += 24.0;}

  double hr = double( UT.time().hour() );
  double mn = double( UT.time().minute() );
  double sc = double( UT.time().second() ) + double ( 0.001 * UT.time().msec() );

  float t2 = (hr + ( mn + sc/60.0)/60.0)*1.002737909;
  double gsth = t1 + t2;

  while (gsth < 0.0) gsth += 24.0;
  while (gsth >24.0) gsth -= 24.0;

  int gh = int(gsth);
  int gm = int((gsth - gh)*60.0);
  int gs = int(((gsth - gh)*60.0 - gm)*60.0);

  QTime gst(gh, gm, gs);

  return gst;
}

QTime KStars::GSTtoLST( QTime GST, float longitude) {
  float lsth = float(GST.hour()) + (float(GST.minute()) +
			   float(GST.second())/60.0)/60.0 + longitude/15.0;

  while (lsth < 0.0) lsth += 24.0;
  while (lsth >24.0) lsth -= 24.0;

  int lh = int(lsth);
  int lm = int((lsth - lh)*60.0);
  int ls = int(((lsth - lh)*60.0 - lm)*60.0);

  QTime lst(lh, lm, ls);

  return lst;
}


//Functions for reading data files
bool KStars::openDataFile( QFile &file, QString s ) {
	QString ResName = "kstars/" + s;
	QString FileName= KDirs->findResource( "data", ResName );
	
	if ( FileName != QString::null ) {
		file.setName( FileName );
		
		if ( !file.open( IO_ReadOnly ) ) {
			QString message = (QString)"Could not open " + FileName;
			KMessageBox::sorry( 0, message, "File open error" );
		}
  } else { //File not found, try the local option
		file.setName( "data/" + s );

		if (!file.open( IO_ReadOnly ) ) { //File not found.  return false.
			QString message =
					(QString)"The file " + s + (QString)" could not be found.\n" +
					(QString)"Place the file in one of the following locations and restart KStars:\n\n" +
					(QString)"$(KDEDIR)/share/apps/kstars/" + s +
					(QString)"\n~/.kde/share/apps/kstars/" + s +
					(QString)"\n$(PATH_TO_KSTARS)/data/" + s;
			KMessageBox::sorry( 0, message, s+" not found!" );
			return( false );
		}													
  }
	return( true );	
}

void KStars::readCLineData( void ) {
	QFile file;
	if ( openDataFile( file, "clines.dat" ) ) {
	  QTextStream stream( &file );

  	while ( !stream.eof() ) {
	  	QString line, name;
		  int rah, ram, ras, dd, dm, ds;
		  QChar sgn;
			QChar *mode;

		  line = stream.readLine();

		  rah = line.mid( 0, 2 ).toInt();
	  	ram = line.mid( 2, 2 ).toInt();
		  ras = line.mid( 4, 2 ).toInt();

		  sgn = line.at( 6 );
	  	dd = line.mid( 7, 2 ).toInt();
		  dm = line.mid( 9, 2 ).toInt();
		  ds = line.mid( 11, 2 ).toInt();
	    mode = new QChar( line.at( 13 ) );
		  hms r( rah, ram, ras );
	  	dms d( dd, dm,  ds );

		  if ( sgn == "-" ) { d.setD( -1.0*d.getD() ); }

		  SkyPoint *o = new SkyPoint( r.getD(), d.getD() );
			o->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
		  clineList.append( o );
			clineModeList.append( mode );
  	}
		file.close();
	}
}

void KStars::readCNameData( void ) {
  QFile file;
	if ( openDataFile( file, "cnames.dat" ) ) {
	  QTextStream stream( &file );

  	while ( !stream.eof() ) {
	  	QString line, name;
		  int rah, ram, ras, dd, dm, ds;
		  QChar sgn;

	  	line = stream.readLine();

		  rah = line.mid( 0, 2 ).toInt();
		  ram = line.mid( 2, 2 ).toInt();
	  	ras = line.mid( 4, 2 ).toInt();

		  sgn = line.at( 6 );
		  dd = line.mid( 7, 2 ).toInt();
	  	dm = line.mid( 9, 2 ).toInt();
		  ds = line.mid( 11, 2 ).toInt();

		  name  = line.mid( 13 ).stripWhiteSpace();
	
		  hms r( rah, ram, ras );
		  dms d( dd, dm,  ds );

	  	if ( sgn == "-" ) { d.setD( -1.0*d.getD() ); }

		  SkyObject *o = new SkyObject( -1, r, d, 0.0, name );
			o->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
	  	cnameList.append( o );
	  }
		file.close();
	}
}

void KStars::readStarData( void ) {
	QFile file;
	if ( openDataFile( file, "sao.dat" ) ) {
	  QTextStream stream( &file );

	  while ( !stream.eof() ) {
		  QString line, name;
	  	float mag;
		  int rah, ram, ras, dd, dm, ds;
		  QChar sgn;

	  	line = stream.readLine();

			name = "star";
		  rah = line.mid( 0, 2 ).toInt();
	  	ram = line.mid( 2, 2 ).toInt();
		  ras = line.mid( 4, 2 ).toInt();
	
		  sgn = line.at( 6 );
	  	dd = line.mid( 7, 2 ).toInt();
		  dm = line.mid( 9, 2 ).toInt();
		  ds = line.mid( 11, 2 ).toInt();

	  	mag = line.mid( 13, 4 ).toFloat();

		  hms r( rah, ram, ras );
		  dms d( dd, dm,  ds );
	
		  if ( sgn == "-" ) { d.setD( -1.0*d.getD() ); }

		  SkyObject *o = new SkyObject( 0, r, d, mag, name );
			o->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
	  	starList.append( o );
			if (mag < 4.0) starCount0 = starList.count();
  	  if (mag < 6.0) starCount1 = starList.count();
			if (mag < 7.0) starCount2 = starList.count();
			if (mag < 7.5) starCount3 = starList.count();
  	}
		file.close();
	}
}

void KStars::readNGCData( void ) {
	QFile file;
	if ( openDataFile( file, "ngc2000.dat" ) ) {
	  QTextStream stream( &file );

  	while ( !stream.eof() ) {
	  	QString line, smag, name, name2, longname;
		  float mag, ramf;
		  int type, ingc, imess, rah, ram, ras, dd, dm;
	  	QChar sgn, cat_id, cat2_id, mflag;

		  line = stream.readLine();

		  cat_id = line.at( 0 ); //check for IC catalog flag
	  	if ( cat_id == " " ) cat_id = 'N'; // if blank, set catalog to NGC
	
		  ingc = line.mid( 1, 4 ).toInt();  // NGC/IC catalog number
			type = line.mid( 6, 1 ).toInt();     // object type identifier
	  	rah = line.mid( 8, 2 ).toInt();
		  ramf = line.mid( 11, 4 ).toFloat();
		  sgn = line.at( 16 );
	  	dd = line.mid( 17, 2 ).toInt();
		  dm = line.mid( 20, 2 ).toInt();
		  smag = line.mid( 23, 4 );
		  if (smag == "    " ) { mag = 99.9; } else { mag = smag.toFloat(); }
			mflag = line.at( 28 );
			if ( mflag=='M' ) {
				cat2_id = cat_id;
				if ( ingc==0 ) cat2_id = ' ';
				cat_id = mflag;
				imess = line.mid( 30, 3 ).toInt();
			}
			longname = line.mid( 34, 30 );
		
		  ram = int( ramf );
		  ras = int( 60.0*( ramf - (float)ram ) );
	
	  	hms r( rah, ram, ras );
		  dms d( dd, dm, 0 );

		  if ( sgn == "-" ) { d.setD( -1.0*d.getD() ); }
	
			QString snum;
			if (cat_id=="I") {
				snum.setNum( ingc );
				name = "IC " + snum;
			} else if (cat_id=="N") {
				snum.setNum( ingc );
				name = "NGC " + snum;
			} else if (cat_id=="M") {
				snum.setNum( imess );
				name = "M " + snum;
				if (cat2_id=="N") {
					snum.setNum( ingc );
					name2 = "NGC " + snum;
				} else if (cat2_id=="I") {
					snum.setNum( ingc );
					name2 = "IC " + snum;
				} else {
					name2 = "";
				}
			}
			
	  	SkyObject *o = new SkyObject( type, r, d, mag, name, name2, longname );
			o->RADecToAltAz( skymap->LSTh.getD(), geo->lat() );
		  if (cat_id=="M") {
	  		messList.append( o );
		  } else if ( cat_id=="N" ) {
			  ngcList.append( o );
			} else if ( cat_id=="I" ) {
				icList.append( o );
			}  	
	  }
		file.close();
	}
}

void KStars::readURLData( void ) {
	QFile file;
	if ( openDataFile( file, "hsturl.dat" ) ) {
	  QTextStream stream( &file );

  	while ( !stream.eof() ) {
	  	QString line = stream.readLine();
			QString name = line.mid( 0, line.find(':') );
			QString url = line.mid( line.find(':')+1 );
		
			switch( name.at(0).unicode() ) {
				case 'M' : //Messier object
					for ( unsigned int i=0; i<messList.count(); ++i ) {
						SkyObject *o = messList.at(i);
						if ( o->name == name ) {
							o->hstURL = url;
							break;
						}
					}
					break;
				case 'N' : //NGC object
					for ( unsigned int i=0; i<ngcList.count(); ++i ) {
						SkyObject *o = ngcList.at(i);
						if ( o->name == name ) {
							o->hstURL = url;
							break;
						}
					}
					break;
				case 'I' : //IC object
					for ( unsigned int i=0; i<icList.count(); ++i ) {
						SkyObject *o = icList.at(i);
						if ( o->name == name ) {
							o->hstURL = url;
							break;
						}
					}
					break;
			}
		}
		file.close();
	}
}

void KStars::readCityData( void ) {
	QFile file;
	if ( openDataFile( file, "Cities.dat" ) ) {
		QTextStream stream( &file );

  	while ( !stream.eof() ) {
	  	QString line, name, state;
			char latsgn, lngsgn;
			int lngD, lngM, lngS;
	  	int latD, latM, latS;
			int TZ;
		  float lng, lat;

		 	line = stream.readLine();

		  name  = line.left( 32 );
	  	state = line.mid( 31, 17 );
		 	latD = line.mid( 49, 2 ).toInt();
		  latM = line.mid( 52, 2 ).toInt();
	  	latS = line.mid( 55, 2 ).toInt();
		 	latsgn = line.at( 58 ).latin1();
		  lngD = line.mid( 60, 3 ).toInt();
	  	lngM = line.mid( 64, 2 ).toInt();
		 	lngS = line.mid( 67, 2 ).toInt();
		  lngsgn = line.at( 70 ).latin1();

	  	lat = (float)latD + ((float)latM + (float)latS/60.0)/60.0;
			lng = (float)lngD + ((float)lngM + (float)lngS/60.0)/60.0;
  	
		 	if ( latsgn == 'S' ) lat *= -1.0;
			if ( lngsgn == 'W' ) lng *= -1.0;
  	
		  TZ = int(lng/-15.0);

			name = name.stripWhiteSpace();
			name = name.left(1).upper() + name.mid(1).lower();
			state = state.stripWhiteSpace();
			state = state.left(1).upper() + state.mid(1).lower();

			GeoLocation *geo = new GeoLocation( lng, lat, name, state, TZ );
			geoList.append( geo );
  	}
		file.close();
	}
}
